//----------------------------------------------------------------------------------------------
// Board Library : Seeed nRF52 Borads 1.1.8
// Board Select  : Seeed nRF52 Borads / Seeed XIAO nRF52840 Sense
// 2024/04/28 @msfujino
//----------------------------------------------------------------------------------------------

#include <bluefruit.h>			        // \Arduino15\packages\Seeeduino\hardware\nrf52\1.1.8\libraries\Bluefruit52Lib\src

#define dataNum 	  8               // receive data number : roll, pitch, yaw, Vbatt (4 data 8 byte)

int ss;                             // RSSI
uint8_t addr[6];                    // device address
bool readingFlag = false;           // data buffer in-use flag
bool notifyFlag = false;            // notify flag
bool connectedFlag = false;         // connected flag
union unionData {
  uint32_t  dataBuff32[dataNum/4];                   // Union for bit convertion
  uint16_t  dataBuff16[dataNum/2];
  uint8_t   dataBuff8[dataNum];
};
union unionData ud;

// Custum "Client" Service and Characteristic
// generated Linux command uuidgen, Example "0dd7eb5a-a8b9-4f45-bcd7-94c674c3b25f"
#define myUUID(val) (const uint8_t[]) { \
    0x5F, 0xB2, 0xC3, 0x74, 0xC6, 0x94, 0xD7, 0xBC, \
    0x45, 0x4F, (uint8_t)(val & 0xff), (uint8_t)(val >> 8), 0x5A, 0xEB, 0xD7, 0x0D }
BLEClientService        myService           (myUUID(0x0000));   // "Client" Service
BLEClientCharacteristic myCharacteristic    (myUUID(0x0030));   // "Client" Characteristtic

// **********************************************************************************************************
void setup()
{
  // initialization of serial port
  Serial.begin(115200);
//  while (!Serial);
  delay(2000);

  // I/O pins settings
  pinMode(LED_RED, OUTPUT);       // LOW:LED ON, HIGH:LED OFF
  pinMode(LED_GREEN, OUTPUT);
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);

  // initialization of Bruefruit
  Bluefruit.configCentralBandwidth(BANDWIDTH_MAX);  
  Bluefruit.begin(0, 1);  // bool begin(uint8_t prph_count, uint8_t central_count)
  Bluefruit.setName("XIAO BLE Central");

  // Custom Service Settings
  myService.begin();
  myCharacteristic.setNotifyCallback(data_notify_callback);  
  myCharacteristic.begin();

  // Blue LED blinking interval setting
  Bluefruit.setConnLedInterval(100);

  // callback functions
  Bluefruit.Central.setDisconnectCallback(disconnect_callback);
  Bluefruit.Central.setConnectCallback(connect_callback);

  // scanner settings
  Bluefruit.Scanner.setRxCallback(scan_callback);
  Bluefruit.Scanner.restartOnDisconnect(true);
  Bluefruit.Scanner.setIntervalMS(100, 50);     // interval:100mS  window:50mS  
  Bluefruit.Scanner.filterUuid(myService.uuid);
  Bluefruit.Scanner.useActiveScan(false);
  Bluefruit.Scanner.start(0);                   // Continuous Scanning

  Serial.println("Initialization end");
}

//******************************************************************************************
void loop() {
  // wait for response from peripheral
  Serial.println("Connecting to Peripheral ........");
  while (!Bluefruit.connected()) {
    Serial.print(".");
    delay(1000);
  }
  Serial.println();
  Serial.println("Connected to Peripheral ........");

  while(notifyFlag == true & Bluefruit.connected())
  {                
    digitalWrite(LED_GREEN, LOW);      
      
    // restoration of received data
    // prohibit event handler access while data buffer is in use          
    readingFlag = true;   
      uint32_t ms = ud.dataBuff32[0];
      uint32_t us = ud.dataBuff32[1];
    readingFlag = false;

    // for serial plotter
    Serial.print(ms); Serial.print(" ");
    Serial.println(us);
  
    digitalWrite(LED_GREEN, HIGH);
    delay(1000);     
  } // while connected and notify
    
  // if disconnected from the peripheral
  digitalWrite(LED_GREEN, HIGH);            // connect indicator OFF
  Serial.print("Disconnected from the peripheral: ");
  for(int i = 5; i > 0; i--) {
    Serial.print(addr[i], HEX); Serial.print(":");
  }
  Serial.println(addr[0]);
} // loop

//*****************************************************************************************************
// Callback invoked when scanner pick up an advertising data
 // \Arduino15\packages\Seeeduino\hardware\nrf52\1.1.8\cores\nRF5\nordic\softdevice\s140_nrf52_7.3.0_API\include\ble_gap.h
 // ble_gap.h:1286 typedef struct { } ble_gap_evt_adv_report_t;
void scan_callback(ble_gap_evt_adv_report_t* report)
{
  Bluefruit.Central.connect(report);
  ss = report->rssi;                        // RSSI
  memcpy(addr, report->peer_addr.addr, 6);  // device address
}

//****************************************************************************************************:
// Callback invoked when an connection is established
void connect_callback(uint16_t conn_handle)
{
  connectedFlag = false;

  Serial.println();  
  if ( !myService.discover(conn_handle) )
  {
    Serial.println("【connect_callback】 Service not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Service");
  
  if ( !myCharacteristic.discover() )
  {
    Serial.println("【connect_callback】 Characteristic not find disconnect!");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 Find Characteristic");
  
  if ( !myCharacteristic.enableNotify() )
  {
    Serial.println("【connect_callback】 Couldn't enable notify for data");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("【connect_callback】 enable notify");

  connectedFlag = true;
}

//**********************************************************************************************************
// Callback invoked when an connection is broken
void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.print("【disconnect_callback】 Disconnected, reason = 0x"); Serial.println(reason, HEX);
}

//*********************************************************************************************************
// Callback invoked when a notifyication is received
void data_notify_callback(BLEClientCharacteristic* chr, uint8_t* data, uint16_t len)
{
//  Serial.print("【notify_callback】 notify_callback, len = "); Serial.println(len);
  while(!(notifyFlag == false | connectedFlag == true)) {
    Serial.println("wait for notifyFlag clear"); // wait for notifyFlag clear
  }
  
  digitalWrite(LED_RED, LOW);

  memcpy(ud.dataBuff8, data, len);
  
  delay(5);   // for LED
  digitalWrite(LED_RED, HIGH);

  notifyFlag = true;
}


